@@ -1,8 +1,11 @@
 package org.jsoup.parser;
 
-import org.junit.Test;
+import org.junit.jupiter.api.Test;
 
-import static org.junit.Assert.*;
+import java.io.BufferedReader;
+import java.io.StringReader;
+
+import static org.junit.jupiter.api.Assertions.*;
 
 /**
  * Test suite for character reader.
@@ -10,6 +13,7 @@ import static org.junit.Assert.*;
  * @author Jonathan Hedley, jonathan@hedley.net
  */
 public class CharacterReaderTest {
+    public final static int maxBufferLen = CharacterReader.maxBufferLen;
 
     @Test public void consume() {
         CharacterReader r = new CharacterReader("one");
@@ -45,20 +49,31 @@ public class CharacterReaderTest {
         assertTrue(r.isEmpty());
 
         assertEquals(CharacterReader.EOF, r.consume());
+        r.unconsume(); // read past, so have to eat again
+        assertTrue(r.isEmpty());
         r.unconsume();
+        assertFalse(r.isEmpty());
+
+        assertEquals('e', r.consume());
+        assertTrue(r.isEmpty());
+
+        assertEquals(CharacterReader.EOF, r.consume());
         assertTrue(r.isEmpty());
-        assertEquals(CharacterReader.EOF, r.current());
     }
 
     @Test public void mark() {
         CharacterReader r = new CharacterReader("one");
         r.consume();
         r.mark();
+        assertEquals(1, r.pos());
         assertEquals('n', r.consume());
         assertEquals('e', r.consume());
         assertTrue(r.isEmpty());
         r.rewindToMark();
+        assertEquals(1, r.pos());
         assertEquals('n', r.consume());
+        assertFalse(r.isEmpty());
+        assertEquals(2, r.pos());
     }
 
     @Test public void consumeToEnd() {
@@ -116,7 +131,15 @@ public class CharacterReaderTest {
         assertEquals('T', r.consume());
         assertEquals("wo ", r.consumeTo("Two"));
         assertEquals('T', r.consume());
-        assertEquals("wo Four", r.consumeTo("Qux"));
+        // To handle strings straddling across buffers, consumeTo() may return the
+        // data in multiple pieces near EOF.
+        StringBuilder builder = new StringBuilder();
+        String part;
+        do {
+            part = r.consumeTo("Qux");
+            builder.append(part);
+        } while (!part.isEmpty());
+        assertEquals("wo Four", builder.toString());
     }
 
     @Test public void advance() {
@@ -165,6 +188,7 @@ public class CharacterReaderTest {
         assertFalse(r.matches("ne Two Three Four"));
         assertEquals("ne Two Three", r.consumeToEnd());
         assertFalse(r.matches("ne"));
+        assertTrue(r.isEmpty());
     }
 
     @Test
@@ -220,10 +244,10 @@ public class CharacterReaderTest {
         assertEquals("Check", two);
         assertEquals("Check", three);
         assertEquals("CHOKE", four);
-        assertTrue(one == two);
-        assertTrue(two == three);
-        assertTrue(three != four);
-        assertTrue(four != five);
+        assertSame(one, two);
+        assertSame(two, three);
+        assertNotSame(three, four);
+        assertNotSame(four, five);
         assertEquals(five, "A string that is longer than 16 chars");
     }
 
@@ -244,5 +268,72 @@ public class CharacterReaderTest {
         assertFalse(r.rangeEquals(18, 5, "CHIKE"));
     }
 
+    @Test
+    public void empty() {
+        CharacterReader r = new CharacterReader("One");
+        assertTrue(r.matchConsume("One"));
+        assertTrue(r.isEmpty());
+
+        r = new CharacterReader("Two");
+        String two = r.consumeToEnd();
+        assertEquals("Two", two);
+    }
+
+    @Test
+    public void consumeToNonexistentEndWhenAtAnd() {
+        CharacterReader r = new CharacterReader("<!");
+        assertTrue(r.matchConsume("<!"));
+        assertTrue(r.isEmpty());
+
+        String after = r.consumeTo('>');
+        assertEquals("", after);
+
+        assertTrue(r.isEmpty());
+    }
+
+    @Test
+    public void notEmptyAtBufferSplitPoint() {
+        CharacterReader r = new CharacterReader(new StringReader("How about now"), 3);
+        assertEquals("How", r.consumeTo(' '));
+        assertFalse(r.isEmpty(), "Should not be empty");
+
+        assertEquals(' ', r.consume());
+        assertFalse(r.isEmpty());
+        assertEquals(4, r.pos());
+        assertEquals('a', r.consume());
+        assertEquals(5, r.pos());
+        assertEquals('b', r.consume());
+        assertEquals('o', r.consume());
+        assertEquals('u', r.consume());
+        assertEquals('t', r.consume());
+        assertEquals(' ', r.consume());
+        assertEquals('n', r.consume());
+        assertEquals('o', r.consume());
+        assertEquals('w', r.consume());
+        assertTrue(r.isEmpty());
+    }
+
+    @Test public void bufferUp() {
+        String note = "HelloThere"; // + ! = 11 chars
+        int loopCount = 64;
+        StringBuilder sb = new StringBuilder();
+        for (int i = 0; i < loopCount; i++) {
+            sb.append(note);
+            sb.append("!");
+        }
+
+        String s = sb.toString();
+        BufferedReader br = new BufferedReader(new StringReader(s));
+
+        CharacterReader r = new CharacterReader(br);
+        for (int i = 0; i < loopCount; i++) {
+            String pull = r.consumeTo('!');
+            assertEquals(note, pull);
+            assertEquals('!', r.current());
+            r.advance();
+        }
+
+        assertTrue(r.isEmpty());
+    }
 
 }
